/* This file is part of Om.  Copyright (C) 2005 Dave Robillard.
 * 
 * Om is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Om is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "AudioOutputNode.h"
#include "InputPort.h"
#include "OutputPort.h"
#include "Plugin.h"
#include "Patch.h"
#include "Om.h"
#include "OSCSender.h"
#include "Connection.h"
#include "JackDriver.h"
#include <cassert>

namespace Om {


AudioOutputNode::AudioOutputNode(const string& path, uint poly, Patch* parent, samplerate srate, size_t buffer_size)
: OutputNode(path, poly, parent, srate, buffer_size),
  m_jack_port(NULL)
{
	m_external_port = new OutputPort(parent, m_name, 0, m_poly, new PortInfo(m_name, AUDIO, OUTPUT), m_buffer_size);
	InputPort* iport = new InputPort(this, "out", 0, m_poly, new PortInfo("out", AUDIO, INPUT), m_buffer_size);
	iport->join(m_external_port);
	
	m_num_ports = 1;
	m_ports.alloc(m_num_ports);
	m_ports.at(0) = iport;

	m_plugin.type(Plugin::Internal);
	m_plugin.lib_path("");
	m_plugin.lib_name("");
	m_plugin.plug_label("audio_output");
	m_plugin.name("Om patch audio output node");

	m_listnode = new ListNode<OutputNode*>(this);

	m_ports.at(0)->set_value(0, 0);
}


AudioOutputNode::~AudioOutputNode()
{
	delete m_external_port;
	delete m_listnode;
	delete m_jack_port;
}


void
AudioOutputNode::activate()
{
	assert(parent_patch() != NULL);
	
	InternalNode::activate();
	if (parent_patch()->parent() == NULL && om != NULL)
		m_jack_port = om->jack_driver()->create_output(parent_patch(), this);
}


void
AudioOutputNode::deactivate()
{
	if (m_is_added)
		remove_from_patch();
	
	InternalNode::deactivate();

	if (parent_patch() != NULL && parent_patch()->parent() == NULL && om != NULL && m_jack_port != NULL)
		om->jack_driver()->unregister_port(m_jack_port->elem());
}


void
AudioOutputNode::add_to_patch()
{
	assert(parent_patch() != NULL);
	
	parent_patch()->output_nodes().push_back(m_listnode);
	InternalNode::add_to_patch();
	
	// Activate jack port now in the audio thread (not before when created, to avoid race issues)
	if (parent_patch()->parent() == NULL) {
		assert(m_jack_port != NULL);
		om->jack_driver()->add_output(m_jack_port);
	}

	m_is_added = true;
}


void
AudioOutputNode::remove_from_patch()
{
	if (m_is_added) {
		if (parent_patch()->parent() == NULL) {
			assert(m_jack_port != NULL);
			m_jack_port = om->jack_driver()->remove_output(m_jack_port->elem());
		}
		ListNode<OutputNode*>* ln = parent_patch()->output_nodes().remove(this);
		om->maid()->push(ln);
		m_listnode = NULL; // avoid a double delete in destructor
		InternalNode::remove_from_patch();
		m_is_added = false;
	}
}


void
AudioOutputNode::send_creation_messages(lo_address addr) const
{
	om->osc_sender()->send_new_port_to(addr, m_external_port);
	InternalNode::send_creation_messages(addr);
}


void
AudioOutputNode::send_deletion_messages(lo_address addr) const
{
	InternalNode::send_deletion_messages(addr);
	om->osc_sender()->send_port_removal_to(addr, m_external_port);
}


void
AudioOutputNode::set_name(const string& new_name)
{
	OutputNode::set_name(new_name);

	if (m_jack_port != NULL)
		jack_port_set_name(m_jack_port->elem()->jack_port(), path().c_str());
}


} // namespace Om

